import sys
import math
import random
import collections
import pygame
import pymunk as pm

import data
import const
import util
import tiletype
import splitter

class GameWindow(object):
    
    def __init__(self):
        pygame.display.set_caption('invisipin')
        self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT))
        try:
            pygame.mixer.init()
            pygame.mixer.music.load(data.filepath('Twisted.mp3'))
            pygame.mixer.music.play(-1)
        except:
            print 'Cannot load music.'
        self.intro()
        # self.game(const.SMALL)
        # self.game(const.LARGE)

    def intro(self):
        next = Intro(self, self.screen).loop()
        if next == 0:
            self.instructions()
        elif next == 1:
            self.game(const.SMALL)
        elif next == 2:
            self.game(const.LARGE)
        elif next == 3:
            sys.exit()
    
    def instructions(self):
        Instructions(self, self.screen).loop()
        self.intro()
    
    def game(self, level):
        Game(self, self.screen, level).loop()
        self.intro()


class Intro(object):

    def __init__(self, window, screen):
        self.window = window
        self.screen = screen

        self.font = pygame.font.Font(data.filepath('Holmes207.ttf'), 48)
        self.buttons = pygame.sprite.RenderUpdates()
        self.create_button('How to play', 300, 0)
        self.create_button('Play small board', 400, 1)
        self.create_button('Play large board', 500, 2)
        self.create_button('Exit game', 600, 3)
        self.muted = False

    def create_button(self, text, posy, bid):
        b = pygame.sprite.Sprite()
        b.image_uns = self.font.render(text, True, (255,255,255))
        b.image_sel = self.font.render(text, True, (128,255,128))
        b.image = b.image_uns
        b.rect = b.image.get_rect()
        b.rect.center = (const.WIDTH / 2, posy)
        b.bid = bid
        self.buttons.add(b)

    def loop(self):
        self.bg = pygame.image.load(data.filepath('bg.png')).convert()
        self.screen.blit(self.bg, (0,0))
        pygame.display.update()

        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        sys.exit()
                    elif event.key == pygame.K_m:
                        if not self.muted:
                            pygame.mixer.music.pause()
                            self.muted = True
                        else:
                            pygame.mixer.music.unpause()
                            self.muted = False
                elif event.type == pygame.MOUSEBUTTONUP:
                    pos = pygame.mouse.get_pos()
                    top = [s for s in self.buttons if s.rect.collidepoint(pos)]
                    if len(top) > 0:
                        b = top[0]
                        return b.bid
                elif event.type == pygame.MOUSEMOTION:
                    pos = pygame.mouse.get_pos()
                    for s in self.buttons:
                        s.image = s.image_uns
                    top = [s for s in self.buttons if s.rect.collidepoint(pos)]
                    if len(top) > 0:
                        top[0].image = top[0].image_sel
            
            self.buttons.clear(self.screen, self.bg)
            self.buttons.draw(self.screen)
            
            pygame.display.flip()


class Instructions(object):

    def __init__(self, window, screen):
        self.window = window
        self.screen = screen
        
        self.muted = False

    def loop(self):
        self.bg = pygame.image.load(data.filepath('bg.png')).convert()
        self.screen.blit(self.bg, (0,0))
        self.bg = pygame.image.load(data.filepath('instructions.png')).convert_alpha()
        self.screen.blit(self.bg, (0,0))
        pygame.display.update()

        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_m:
                        if not self.muted:
                            pygame.mixer.music.pause()
                            self.muted = True
                        else:
                            pygame.mixer.music.unpause()
                            self.muted = False
                    else:
                        self.done = True
                elif event.type == pygame.MOUSEBUTTONUP:
                    self.done = True

            pygame.display.flip()


class Game(object):

    def __init__(self, window, screen, levelsize):
        self.window = window
        self.screen = screen
        self.clock = pygame.time.Clock()

        self.paused = False
        self.muted = False

        self.font = pygame.font.Font(data.filepath('Holmes207.ttf'), 36)

        self.space = pm.Space()
        self.space.gravity = pm.Vec2d(0, -1200)
        
        self.level = LevelGrid(self, self.space, levelsize['size'], levelsize['types'], levelsize['probs'])
        self.balls = BallManager(self.space, self.level)
        self.pieces = PuzzlePieceManager(self.level)
        self.breakers = BreakerManager(self.level)
        self.breakers.add_breakers(levelsize['nbreakers'])
        self.win_msg = pygame.sprite.RenderUpdates()
        self.piece_held = None
        
        self.level_visible = False
        self.won = False

        pygame.font.init()
    
    def loop(self):
        self.bg = pygame.image.load(data.filepath('bg.png')).convert()
        self.screen.blit(self.bg, (0,0))
        pygame.display.update()

        grid_img = pygame.sprite.RenderUpdates()
        grid_img.add(self.level.get_grid_image())
        
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.MOUSEMOTION:
                    self.on_mousemotion(event)
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    self.on_mousedown(event)
                elif event.type == pygame.MOUSEBUTTONUP:
                    self.on_mouseup(event)

            # clear
            self.level.clear(self.screen, self.bg)
            self.level.indicators_con.clear(self.screen, self.bg)
            self.pieces.clear(self.screen, self.bg)
            self.balls.clear(self.screen, self.bg)
            self.breakers.clear(self.screen, self.bg)
            self.win_msg.clear(self.screen, self.bg)

            dt = self.clock.tick(const.FPS) / 1000.0

            # update
            self.space.step(1/90.0)
            self.level.update(dt)
            self.level.indicators_con.update(dt)
            self.pieces.update(dt)
            self.breakers.update(dt)
            self.balls.update(dt)

            # draw
            grid_img.draw(self.screen)
            if self.level_visible:
                self.level.draw(self.screen)
            self.level.indicators_con.draw(self.screen)
            self.pieces.draw(self.screen)
            self.balls.draw(self.screen)
            self.breakers.draw(self.screen)
            self.win_msg.draw(self.screen)

            pygame.display.flip()

    def pause(self):
        self.paused = not self.paused

    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            self.done = True
        elif event.key == pygame.K_p:
            self.pause()
        # elif event.key == pygame.K_v:
        #     self.level_visible = not self.level_visible
        # elif event.key == pygame.K_w:
        #     self.win()
        elif event.key == pygame.K_c:
            self.balls.remove_all()
        elif event.key == pygame.K_m:
            if not self.muted:
                pygame.mixer.music.pause()
                self.muted = True
            else:
                pygame.mixer.music.unpause()
                self.muted = False

    def on_mousedown(self, event):
        if self.piece_held:
            self.piece_held.release()
            self.piece_held = None
        else:
            pos = pygame.mouse.get_pos()
            clicked = [s for s in self.breakers if s.rect.collidepoint(pos)]
            clicked += [s for s in self.pieces if s.rect.collidepoint(pos)]
            if len(clicked) > 0:
                clicked[0].hold()
                self.piece_held = clicked[0]
    
    def on_mouseup(self, event):
        if self.piece_held:
            self.piece_held.release()
            self.piece_held = None
        else:
            pos = pygame.mouse.get_pos()
            if pos[1] <= 200 and pos[0] >= self.level.pos[0] and pos[0] < self.level.pos[0] + self.level.tilesize * self.level.sizex:
                self.balls.add_ball(pos)
            if self.won:
                clicked = [s for s in self.win_msg if s.rect.collidepoint(pos)]
                if len(clicked) > 0:
                    self.done = True

    def on_mousemotion(self, event):
        if self.piece_held:
            pos = pygame.mouse.get_pos()
            if self.level.pos_on_grid(pos):
                self.level.light_up_tilepos(self.piece_held.get_all_tilepos())
            else:
                self.level.light_up_tilepos([])
    
    def almost(self):
        almost = pygame.sprite.Sprite()
        almost.image = pygame.image.load(data.filepath('almost.png')).convert_alpha()
        almost.rect = almost.image.get_rect()
        almost.rect.bottomright = (const.WIDTH - 10, const.HEIGHT - 10)
        self.win_msg.add(almost)

    def win(self):
        if not self.won:
            self.won = True
            win = pygame.sprite.Sprite()
            win.image = pygame.image.load(data.filepath('win.png')).convert_alpha()
            win.rect = win.image.get_rect()
            win.rect.bottomright = (const.WIDTH - 10, const.HEIGHT - 10)
            self.win_msg.add(win)
            self.balls.win_animation = True


class Ball(pygame.sprite.Sprite):

    def __init__(self, pos, space, level):
        pygame.sprite.Sprite.__init__(self)
        self.pos = list(pos)
        self.space = space
        self.level = level
        self.radius = 6
        # self.image = pygame.surface.Surface((self.radius * 2, self.radius * 2))
        # pygame.draw.circle(self.image, (255,0,0), (self.radius,self.radius), self.radius, 2)
        self.image = pygame.image.load(data.filepath('ball.png')).convert_alpha()
        self.rect = self.image.get_rect()
        self.rect.center = self.pos

        self.mass = 10
        self.moment = pm.moment_for_circle(self.mass, 0, self.radius)
        self.body = pm.Body(self.mass, self.moment)
        self.body.position = pos[0], util.flip(pos[1])
        self.shape = pm.Circle(self.body, self.radius)
        self.shape.friction = 0.9
        self.shape.elasticity = 0.3
        self.space.add(self.body, self.shape)
        
        self.done = False
    
    def remove_shape(self):
        self.space.remove(self.shape)
    
    def update(self, dt):
        self.pos[0] = self.body.position[0]
        self.pos[1] = util.flip(self.body.position[1])
        self.rect.center = self.pos
        if self.rect.top > const.HEIGHT:
            self.done = True
            self.remove_shape()
        while self.rect.centerx > self.level.pos[0] + self.level.tilesize * self.level.sizex:
            self.rect.centerx -= self.level.tilesize * self.level.sizex
            self.body.position = (self.body.position[0] - self.level.tilesize * self.level.sizex, self.body.position[1])
        while self.rect.centerx < self.level.pos[0]:
            self.rect.centerx += self.level.tilesize * self.level.sizex
            self.body.position = (self.body.position[0] + self.level.tilesize * self.level.sizex, self.body.position[1])


class BallManager(pygame.sprite.RenderUpdates):
    
    def __init__(self, space, level):
        pygame.sprite.RenderUpdates.__init__(self)
        self.space = space
        self.level = level
        self.new_ball_timer = 0
        self.timer_between_balls = 0.5
        self.win_animation = False
        self.win_anim_x = 0
    
    def add_ball(self, pos):
        if self.new_ball_timer <= 0:
            self.add(Ball(pos, self.space, self.level))
            self.new_ball_timer = self.timer_between_balls
            return True
        return False
    
    def remove_all(self):
        for ball in self:
            ball.remove_shape()
        self.empty()

    def update(self, dt):
        if self.new_ball_timer > 0:
            self.new_ball_timer -= dt
        for ball in self:
            ball.update(dt)
            if ball.done:
                self.remove(ball)
        if self.win_animation:
            self.timer_between_balls = 0.2
            ok = self.add_ball((self.level.pos[0] + 10 + self.win_anim_x, 120))
            if ok:
                self.win_anim_x += 20
                if self.win_anim_x >= self.level.tilesize * self.level.sizex:
                    self.win_anim_x = 0


class LevelTile(pygame.sprite.Sprite):
    
    def __init__(self, pos, size, space, neighbor_types, types, probs):
        pygame.sprite.Sprite.__init__(self)
        self.space = space
        self.pos = list(pos)
        
        allowed_types = list(types)
        for p in tiletype.PROHIBITED_PAIRS:
            if neighbor_types[0] == p[0] and p[1] in allowed_types:
                allowed_types.remove(p[1])
            if neighbor_types[1] == p[1] and p[0] in allowed_types:
                allowed_types.remove(p[0])
        self.type = tiletype.get_random(allowed_types, probs)
        self.image, self.shapes = self.type(self.pos, size)
        for shape in self.shapes:
            self.space.add(shape)

        self.rect = self.image.get_rect()
        self.rect.topleft = self.pos
    
    def disable(self):
        for shape in self.shapes:
            shape.sensor = True
    
    def enable(self):
        for shape in self.shapes:
            shape.sensor = False


class LevelGrid(pygame.sprite.RenderUpdates):
    
    def __init__(self, game, space, size, types, probs):
        pygame.sprite.RenderUpdates.__init__(self)
        self.game = game
        self.space = space
        self.tilesize = 56
        self.sizex = size[0]
        self.sizey = size[1]
        self.pos = (const.WIDTH/2 - (self.tilesize * self.sizex)/2, 200)
        self.tiles = [[None for y in range(self.sizey)] for x in range(self.sizex)]
        self.locked_tiles = [[None for y in range(self.sizey)] for x in range(self.sizex)]
        self.locked_breakers = [[None for y in range(self.sizey)] for x in range(self.sizex)]
        self.init_indicators()
        
        for x in range(self.sizex):
            for y in range(self.sizey):
                tpos = (self.pos[0] + x * self.tilesize,
                        self.pos[1] + y * self.tilesize)
                neighbor_types = []
                if x == 0:
                    neighbor_types.append(tiletype.empty)
                else:
                    neighbor_types.append(self.tiles[x-1][y].type)
                if x == self.sizex - 1:
                    neighbor_types.append(self.tiles[0][y].type)
                else:
                    neighbor_types.append(tiletype.empty)
                tile = LevelTile(tpos, self.tilesize, self.space, neighbor_types, types, probs)
                self.add(tile)
                self.tiles[x][y] = tile
    
    def init_indicators(self):
        self.indicators = [[pygame.sprite.Sprite() for y in range(self.sizey)] for x in range(self.sizex)]
        self.indicators_con = pygame.sprite.RenderUpdates()
        self.ind_alpha = 64
        for x in range(self.sizex):
            for y in range(self.sizey):
                ind = self.indicators[x][y]
                self.indicators_con.add(ind)
                ind.image = pygame.surface.Surface((self.tilesize, self.tilesize))
                ind.image.fill((255,255,255))
                ind.image.set_alpha(0)
                ind.rect = ind.image.get_rect()
                ind.rect.topleft = (self.pos[0] + x * self.tilesize,
                                    self.pos[1] + y * self.tilesize)
    
    def pos_on_grid(self, pos):
        return (pos[0] >= self.pos[0]
            and pos[0] < self.pos[0] + self.sizex * self.tilesize
            and pos[1] >= self.pos[1]
            and pos[1] < self.pos[1] + self.sizey * self.tilesize)
    
    def tilepos(self, pos):
        return ((pos[0] - self.pos[0]) / self.tilesize,
                (pos[1] - self.pos[1]) / self.tilesize)
    
    def get_grid_image(self):
        grid_img = pygame.surface.Surface((self.sizex * self.tilesize, self.sizey * self.tilesize))
        grid_img.set_alpha(96)
        for x in range(1, self.sizex):
            pygame.draw.lines(grid_img, (141,118,40), False, [(x * self.tilesize,0), (x * self.tilesize, self.sizey * self.tilesize)], 1)
        for y in range(1, self.sizey):
            pygame.draw.lines(grid_img, (141,118,40), False, [(0,y * self.tilesize), (self.sizey * self.tilesize, y * self.tilesize)], 1)
        grid = pygame.sprite.Sprite()
        grid.image = grid_img
        grid.rect = grid.image.get_rect()
        grid.rect.topleft = self.pos
        return grid
    
    def tile_grid(self):
        return splitter.split_grid(self.sizex, self.sizey)[1]
    
    def check_feasible_tilepos(self, tileposlist):
        for pos in tileposlist:
            if pos[0] < 0 or pos[0] >= self.sizex or pos[1] < 0 or pos[1] >= self.sizey:
                return False
            if self.locked_tiles[pos[0]][pos[1]] != None:
                return False
        return True
    
    def light_up_tilepos(self, tileposlist):
        if not self.check_feasible_tilepos(tileposlist):
            tileposlist = []
        for x in range(self.sizex):
            for y in range(self.sizey):
                if (x,y) in tileposlist:
                    self.indicators[x][y].image.set_alpha(self.ind_alpha)
                else:
                    self.indicators[x][y].image.set_alpha(0)
    
    def lock_tiles(self, tilelist):
        # Assume feasible
        for tile in tilelist:
            tpos = self.tilepos(tile.rect.center)
            self.locked_tiles[tpos[0]][tpos[1]] = tile
        ret = self.check_victory()
        if ret == 0:
            self.game.almost()
        elif ret == 1:
            self.game.win()

    def unlock_tiles(self, tilelist):
        for x in range(self.sizex):
            for y in range(self.sizey):
                if self.locked_tiles[x][y] in tilelist:
                    self.locked_tiles[x][y] = None
        if not self.game.won:
            self.game.win_msg.empty()

    # -1 for not done yet, 0 for incorrect full board, 1 for correct full board
    def check_victory(self):
        correct = True
        for x in range(self.sizex):
            for y in range(self.sizey):
                if self.locked_tiles[x][y] == None:
                    return -1
                if self.locked_tiles[x][y].type != self.tiles[x][y].type:
                    correct = False
        return correct
                
    def check_feasible_breaker(self, breakerpos):
        if breakerpos[0] < 0 or breakerpos[0] >= self.sizex or breakerpos[1] < 0 or breakerpos[1] >= self.sizey:
            return False
        if self.locked_breakers[breakerpos[0]][breakerpos[1]] != None:
            return False
        return True

    def lock_breaker(self, breaker):
        # Assume feasible
        tpos = self.tilepos(breaker.rect.center)
        self.locked_breakers[tpos[0]][tpos[1]] = breaker
        self.tiles[tpos[0]][tpos[1]].disable()

    def unlock_breaker(self, breaker):
        for x in range(self.sizex):
            for y in range(self.sizey):
                if self.locked_breakers[x][y] == breaker:
                    self.locked_breakers[x][y] = None
                    self.tiles[x][y].enable()
                    break


class PuzzlePieceManager(pygame.sprite.LayeredUpdates):
    
    def __init__(self, level):
        pygame.sprite.LayeredUpdates.__init__(self)
        self.level = level
        self.tilepieces = self.level.tile_grid()
        for tilepiece in self.tilepieces:
            if random.randint(1, 2) == 1:
                posx = random.randint(40, 200)
                posy = random.randint(50, const.HEIGHT - 200)
            else:
                posx = random.randint(const.WIDTH - 200, const.WIDTH - 40)
                posy = random.randint(50, const.HEIGHT - 50)
            offsetx = tilepiece[0][0] * self.level.tilesize
            offsety = tilepiece[0][1] * self.level.tilesize
            ptiles = []
            for tile in tilepiece:
                x = tile[0]
                y = tile[1]
                ptile = PuzzlePieceTile((posx + x * self.level.tilesize - offsetx,
                                         posy + y * self.level.tilesize - offsety),
                                        self.level.tiles[x][y],
                                        self.level)
                self.add(ptile)
                ptiles.append(ptile)
            for ptile in ptiles:
                ptile.others = [(p, (p.rect.center[0] - ptile.rect.center[0],
                                     p.rect.center[1] - ptile.rect.center[1]))
                                     for p in ptiles if p != ptile]
    
    def update(self, dt):
        for t in self:
            t.update(dt)
            if t.move_to_front:
                self.move_to_front(t)
                t.move_to_front = False


class PuzzlePieceTile(pygame.sprite.Sprite):
    
    def __init__(self, pos, tile, level):
        pygame.sprite.Sprite.__init__(self)
        self.level = level
        self.pos = list(pos)
        self.image = tile.image
        self.type = tile.type
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
        self.border_color = (247, 240, 185)
        pygame.draw.lines(self.image, self.border_color, False, [(0,0), (self.rect.width-1,0)], 1)
        pygame.draw.lines(self.image, self.border_color, False, [(0,self.rect.height-1), (self.rect.width-1,self.rect.height-1)], 1)
        pygame.draw.lines(self.image, self.border_color, False, [(0,0), (0,self.rect.height-1)], 1)
        pygame.draw.lines(self.image, self.border_color, False, [(self.rect.width-1,0), (self.rect.width-1,self.rect.height-1)], 1)
        self.dragged = False
        self.pos_before_drag = self.pos
        self.others = [] # set by manager
        self.move_to_front = False
        
    def hold(self):
        self.dragged = True
        self.pos_before_drag = self.pos
        self.move_to_front = True
        for other in self.others:
            other[0].move_to_front = True
        self.level.unlock_tiles([self] + [o[0] for o in self.others])

    def release(self):
        self.dragged = False
        if self.level.pos_on_grid(self.rect.center):
            all_tilepos = self.get_all_tilepos()
            tilepos = all_tilepos[0]
            if self.level.check_feasible_tilepos(all_tilepos):
                self.rect.topleft = (self.level.pos[0] + tilepos[0] * self.level.tilesize,
                    self.level.pos[1] + tilepos[1] * self.level.tilesize)
                self.level.lock_tiles([self] + [o[0] for o in self.others])
            else:
                self.fail_drag()
        else:
            # Do not allow other tiles to overlap grid
            for other in self.others:
                if self.level.pos_on_grid(other[0].rect.center):
                    self.fail_drag()
                    break
        self.bring_others()
        self.level.light_up_tilepos([])
        self.pos = self.rect.center
    
    def get_all_tilepos(self):
        tilepos = self.level.tilepos(self.rect.center)
        all_tilepos = [tilepos]
        for other in self.others:
            all_tilepos.append(self.level.tilepos(other[0].rect.center))
        return all_tilepos
    
    def fail_drag(self):
        self.pos = self.pos_before_drag
        self.rect.center = self.pos

    def bring_others(self):
        for other in self.others:
            other[0].rect.center = (self.rect.center[0] + other[1][0],
                                    self.rect.center[1] + other[1][1])

    def update(self, dt):
        if self.dragged:
            self.rect.center = pygame.mouse.get_pos()
            self.bring_others()


class BreakerManager(pygame.sprite.RenderUpdates):
    
    def __init__(self, level):
        pygame.sprite.RenderUpdates.__init__(self)
        self.level = level
    
    def add_breakers(self, nbreakers):
        for i in range(nbreakers):
            self.add(Breaker((60 + 70 * i, 660), self.level))


class Breaker(pygame.sprite.Sprite):
    
    def __init__(self, pos, level):
        pygame.sprite.Sprite.__init__(self)
        self.level = level
        self.pos = list(pos)
        self.image = pygame.image.load(data.filepath('breaker.png')).convert_alpha()
        
        self.rect = self.image.get_rect()
        self.rect.center = self.pos
        self.dragged = False
        self.pos_before_drag = self.pos
        
    def hold(self):
        self.dragged = True
        self.pos_before_drag = self.pos
        self.level.unlock_breaker(self)

    def release(self):
        self.dragged = False
        if self.level.pos_on_grid(self.rect.center):
            tilepos = self.get_all_tilepos()[0]
            if self.level.check_feasible_breaker(tilepos):
                self.rect.topleft = (self.level.pos[0] + tilepos[0] * self.level.tilesize,
                    self.level.pos[1] + tilepos[1] * self.level.tilesize)
                self.level.lock_breaker(self)
            else:
                self.fail_drag()
        self.level.light_up_tilepos([])
        self.pos = self.rect.center

    def fail_drag(self):
        self.pos = self.pos_before_drag
        self.rect.center = self.pos

    def get_all_tilepos(self):
        tilepos = self.level.tilepos(self.rect.center)
        return [tilepos]

    def update(self, dt):
        if self.dragged:
            self.rect.center = pygame.mouse.get_pos()
