import pymunk as pm
import pygame
import random
import util
import data

def make_circle(image, img_size, img_pos, pos, radius):
    pygame.draw.circle(image, (0,255,0), (int(img_size * pos[0]), int(img_size * pos[1])), radius, 2)
    body = pm.Body()
    body.position = img_pos[0] + img_size * pos[0], util.flip(img_pos[1] + img_size * pos[1])
    shape = pm.Circle(body, radius)
    shape.friction = 0.99
    shape.elasticity = 0.3
    return shape

def empty(pos, size):
    # return pygame.surface.Surface((size, size)), []
    image = pygame.image.load(data.filepath('empty.png')).convert_alpha()
    return image, []

def dot(pos, size):
    radius = 6
    # image = pygame.surface.Surface((size, size))
    # pygame.draw.circle(image, (0,255,0), (size/2, size/2), radius, 2)
    image = pygame.image.load(data.filepath('dot.png')).convert_alpha()
    body = pm.Body()
    body.position = pos[0] + size/2, util.flip(pos[1] + size/2)
    shape = pm.Circle(body, radius)
    shape.friction = 0.99
    return image, [shape]

def dot3(pos, size):
    shapes = []
    image = pygame.surface.Surface((size, size))
    radius = 3
    shapes.append(make_circle(image, size, pos, (0.25, 0.25), radius))
    shapes.append(make_circle(image, size, pos, (0.75, 0.25), radius))
    shapes.append(make_circle(image, size, pos, (0.5, 0.75), radius))
    image = pygame.image.load(data.filepath('dot3.png')).convert_alpha()    
    return image, shapes

def dot4(pos, size):
    shapes = []
    image = pygame.surface.Surface((size, size))
    radius = 3
    shapes.append(make_circle(image, size, pos, (0.25, 0.25), radius))
    shapes.append(make_circle(image, size, pos, (0.25, 0.75), radius))
    shapes.append(make_circle(image, size, pos, (0.75, 0.25), radius))
    shapes.append(make_circle(image, size, pos, (0.75, 0.75), radius))
    image = pygame.image.load(data.filepath('dot4.png')).convert_alpha()    
    return image, shapes

def ramp_left(pos, size):
    # image = pygame.surface.Surface((size, size))
    # pygame.draw.lines(image, (0,255,0), False, [(size,0), (0,size)], 2)
    image = pygame.image.load(data.filepath('ramp_left.png')).convert_alpha()
    body = pm.Body()
    frompos = pm.Vec2d(pos[0] + size, util.flip(pos[1]))
    topos = pm.Vec2d(pos[0], util.flip(pos[1] + size))
    shape = pm.Segment(body, frompos, topos, 3)
    shape.friction = 0.99
    return image, [shape]

def ramp_right(pos, size):
    # image = pygame.surface.Surface((size, size))
    # pygame.draw.lines(image, (0,255,0), False, [(0,0), (size,size)], 2)
    image = pygame.image.load(data.filepath('ramp_right.png')).convert_alpha()
    body = pm.Body()
    frompos = pm.Vec2d(pos[0], util.flip(pos[1]))
    topos = pm.Vec2d(pos[0] + size, util.flip(pos[1] + size))
    shape = pm.Segment(body, frompos, topos, 3)
    shape.friction = 0.99
    return image, [shape]

def vertical(pos, size):
    # image = pygame.surface.Surface((size, size))
    # pygame.draw.lines(image, (0,255,0), False, [(size/2,0), (size/2,size)], 2)
    image = pygame.image.load(data.filepath('vertical.png')).convert_alpha()
    body = pm.Body()
    frompos = pm.Vec2d(pos[0] + size/2, util.flip(pos[1]))
    topos = pm.Vec2d(pos[0] + size/2, util.flip(pos[1] + size))
    shape = pm.Segment(body, frompos, topos, 3)
    shape.friction = 0.99
    return image, [shape]

def curved_ramp_left(pos, size):
    image = pygame.image.load(data.filepath('curved_ramp_left.png')).convert_alpha()
    body = pm.Body()
    shapes = []
    frompos = pm.Vec2d(pos[0], util.flip(pos[1] + size))
    topos = pm.Vec2d(pos[0] + 16.5, util.flip(pos[1] + size - 3))
    shapes.append(pm.Segment(body, frompos, topos, 3))
    topos2 = pm.Vec2d(pos[0] + 31, util.flip(pos[1] + size - 6.5))
    shapes.append(pm.Segment(body, topos, topos2, 3))
    topos3 = pm.Vec2d(pos[0] + 45, util.flip(pos[1] + size - 17))
    shapes.append(pm.Segment(body, topos2, topos3, 3))
    topos4 = pm.Vec2d(pos[0] + 52, util.flip(pos[1] + size - 34.5))
    shapes.append(pm.Segment(body, topos3, topos4, 3))
    topos5 = pm.Vec2d(pos[0] + size, util.flip(pos[1]))
    shapes.append(pm.Segment(body, topos4, topos5, 3))
    for shape in shapes:
        shape.friction = 0.99
    return image, shapes

def curved_ramp_right(pos, size):
    image = pygame.image.load(data.filepath('curved_ramp_right.png')).convert_alpha()
    body = pm.Body()
    shapes = []
    frompos = pm.Vec2d(pos[0] + size, util.flip(pos[1] + size))
    topos = pm.Vec2d(pos[0] + size - 16.5, util.flip(pos[1] + size - 3))
    shapes.append(pm.Segment(body, frompos, topos, 3))
    topos2 = pm.Vec2d(pos[0] + size - 31, util.flip(pos[1] + size - 6.5))
    shapes.append(pm.Segment(body, topos, topos2, 3))
    topos3 = pm.Vec2d(pos[0] + size - 45, util.flip(pos[1] + size - 17))
    shapes.append(pm.Segment(body, topos2, topos3, 3))
    topos4 = pm.Vec2d(pos[0] + size - 52, util.flip(pos[1] + size - 34.5))
    shapes.append(pm.Segment(body, topos3, topos4, 3))
    topos5 = pm.Vec2d(pos[0], util.flip(pos[1]))
    shapes.append(pm.Segment(body, topos4, topos5, 3))
    for shape in shapes:
        shape.friction = 0.99
    return image, shapes

# ALLTYPES = [empty, dot, ramp_left, ramp_right]
# PROBABILITIES = {empty: 0.2, dot: 0.2, ramp_left: 0.3, ramp_right: 0.3}
ALLTYPES = [empty, dot, dot3, dot4, ramp_left, ramp_right, vertical]
PROBABILITIES = {empty: 0.05, dot: 0.15, dot3: 0.15, dot4: 0.15, ramp_left: 0.2, ramp_right: 0.2, vertical: 0.10}
PROHIBITED_PAIRS = [(ramp_right, ramp_left), (dot4, ramp_left), (ramp_right, dot4),
    (curved_ramp_right, ramp_left), (ramp_right, curved_ramp_left),
    (curved_ramp_right, curved_ramp_left),
    (dot4, curved_ramp_left),
    (curved_ramp_right, dot4)]

ALLTYPES_LARGE = [empty, dot, dot3, dot4, ramp_left, ramp_right, vertical, curved_ramp_left, curved_ramp_right]
PROBABILITIES_LARGE = {empty: 0.05, dot: 0.11, dot3: 0.115, dot4: 0.115, ramp_left: 0.15, ramp_right: 0.15, vertical: 0.10, curved_ramp_left: 0.10, curved_ramp_right: 0.10}


def get_random(types, probs):
    r = random.uniform(0, 1)
    sp = float(sum([probs[t] for t in types]))
    prob = 0.0
    for t in types:
        prob += probs[t] / sp
        if r < prob:
            return t
    assert(False)
    return None
